// $Id: ModuleCreator.java,v 1.10 2004/06/14 20:59:05 m31 Exp $
//
// This file is part of the program suite "Principia Mathematica II"
// which is a working prototype for the main project:
// "Hilbert II" - http://www.qedeq.org
//
// Copyright 2001-2004  Michael Meyling <mime@qedeq.org>.
//
// "Principia Mathematica II" is free software; you can redistribute
// it and/or modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.

package com.meyling.principia.module;


import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.meyling.principia.argument.Argument;
import com.meyling.principia.argument.ArgumentException;
import com.meyling.principia.argument.Counter;
import com.meyling.principia.argument.Text;
import com.meyling.principia.io.Output;
import com.meyling.principia.io.ParsingException;
import com.meyling.principia.io.ParsingTable;
import com.meyling.principia.io.Position;
import com.meyling.principia.io.TextInput;
import com.meyling.principia.io.Utility;
import com.meyling.principia.logic.paragraph.ParagraphCreator;


/**
 * This class provides static methods for reading and
 * writing an {@link Module} and its subcomponents.
 *
 * @version $Revision: 1.10 $
 * @author  Michael Meyling
 */
public final class ModuleCreator {

    // now follow the strings that represent operators

    /** string for a {@link ParagraphList} */
    public static final String MODULE_PARAGRAPH_LIST
        = "PARAGRAPHS";

    /** string for a {@link Module} */
    public static final String MODULE_MODULE
        = "MODULE";

    /** string for a {@link ImportList} */
    public static final String MODULE_IMPORT_LIST
        = "IMPORTS";

    /** string for a {@link Import} */
    public static final String MODULE_IMPORT
        = "IMPORT";

    /** string for a {@link Name} */
    public static final String MODULE_NAME
        = "NAME";

    /** string for a {@link Version} */
    public static final String MODULE_VERSION
        = "VERSION";

    /** string for a {@link LocationList} */
    public static final String MODULE_LOCATION_LIST
        = "LOCATIONS";

    /** string for a {@link Location} */
    public static final String MODULE_LOCATION
        = "LOCATION";

    /** string for a {@link Specification} */
    public static final String MODULE_SPECIFICATION
        = "SPEC";

    /** string for a {@link Header} */
    public static final String MODULE_HEADER
        = "HEADER";

    /** string for a {@link Description} */
    public static final String MODULE_DESCRIPTION
        = "DESCRIPTION";

    /** string for a {@link Headline} */
    public static final String MODULE_HEADLINE
        = "HEADLINE";

    /** string for a {@link Email} */
    public static final String MODULE_EMAIL
        = "EMAIL";

    /** string for a {@link AuthorList} */
    public static final String MODULE_AUTHOR_LIST
        = "AUTHORS";

    /** string for a {@link Author} */
    public static final String MODULE_AUTHOR
        = "AUTHOR";

    /** string for a {@link UsedbyList} */
    public static final String MODULE_USEDBY_LIST
        = "USEDBY";


    /**
     * for converting a <code>String</code> into an
     * (Argument) <code>class</code>
     */
    private final static Map string2class = new HashMap();

    /**
     * for converting a (Argument) <code>class</code> into an
     * <code>String</code>
     */
    private final static Map class2string = new HashMap();

    /**
     * for converting an created Argument into an input file position
     */
    private final ParsingTable parsing = new ParsingTable();

    /**
     * load from this source
     */
    private final TextInput input;



    /**
     * Make attribut initialization.
     */
    static {
        // fill hash map
        string2class.put(MODULE_PARAGRAPH_LIST,
                         ParagraphList.class);
        string2class.put(MODULE_MODULE,
                         Module.class);
        string2class.put(MODULE_IMPORT_LIST,
                         ImportList.class);
        string2class.put(MODULE_IMPORT,
                         Import.class);
        string2class.put(MODULE_NAME,
                         Name.class);
        string2class.put(MODULE_VERSION,
                         Version.class);
        string2class.put(MODULE_LOCATION_LIST,
                         LocationList.class);
        string2class.put(MODULE_LOCATION,
                         Location.class);
        string2class.put(MODULE_SPECIFICATION,
                         Specification.class);
        string2class.put(MODULE_HEADER,
                         Header.class);
        string2class.put(MODULE_DESCRIPTION,
                         Description.class);
        string2class.put(MODULE_HEADLINE,
                         Headline.class);
        string2class.put(MODULE_EMAIL,
                         Email.class);
        string2class.put(MODULE_AUTHOR_LIST,
                         AuthorList.class);
        string2class.put(MODULE_AUTHOR,
                         Author.class);
        string2class.put(MODULE_USEDBY_LIST,
                         UsedbyList.class);

        // add all the basic staff from package

        //  com.meyling.principia.logic.paragraph
        string2class.putAll(ParagraphCreator.getMap());
        // build inverse mapping
        for (Iterator it = string2class.entrySet().iterator(); it.hasNext(); ) {
            Map.Entry entry = (Map.Entry) it.next();
            class2string.put(entry.getValue(), entry.getKey());
        }

    }


    /**
     * Constructs a new module creator.
     *
     * @param   input   read from here
     */
    public ModuleCreator(final TextInput input) {
        this.input = input;
    }



    /**
     * Get the mapping of Strings to operator classes. The returned Map
     * should not be modified.
     *
     * @return  mapping of known Strings, do not change this Map
     */
    public final static Map getMap() {
        return string2class;
    }


    /**
     * Get name of known operator class.
     *
     * @param   clazz   look for name of this class
     * @return  name of operator class
     * @throws  IllegalArgumentException    if operator class
     *          <code>clazz</code>
     *          is unknown
     */
    public final static String getName(final Class clazz)
            throws IllegalArgumentException {
        final String result;
        try {
            result = (String) class2string.get(clazz);
        } catch (Exception e) {
            throw new IllegalArgumentException(e.toString());
        }
        if (result == null) {
            throw new IllegalArgumentException(clazz.getName()
                + ModuleConstants.UNKNOWN_CLASS);
        }
        return result;
    }


    /**
     * Get class for known name of operator.
     *
     * @param   name   look for class with this name
     * @return  class with this name
     * @throws  IllegalArgumentException    if no class with this name
     *          known
     */
    public final static Class getClass(final String name)
            throws IllegalArgumentException {
        final Class clazz;
        try {
            clazz = (Class) string2class.get(name);
        } catch (Exception e) {
            throw new IllegalArgumentException(e.toString());
        }
        if (clazz == null) {
            throw new IllegalArgumentException(
                ModuleConstants.UNKNOWN_OPERATOR
                + Utility.quote(name));
        }
        return clazz;
    }


    /**
     * Creates a new {@link Argument}. Could't create a new
     * {@link Counter}
     * or a {@link Text}.
     *
     * @param   kind        class to generate
     * @param   argument    argument for constructor of that class
     * @return  new generated Argument
     * @throws  ArgumentException   if creation failed
     * @throws  IllegalArgumentException   if creation failed
     *           because of a programming error
     */
    public final static Argument createArgument(final Class kind,
            final Argument[] argument)
            throws IllegalArgumentException, ArgumentException {

        try {
            Constructor constructor = kind.getConstructor(new Class[] {
                Argument[].class});
            if (constructor == null) {
                throw new IllegalArgumentException(kind.getName() 
                + "couldn't be created with an argument list");
            }
            return (Argument) constructor.newInstance(
                new Object[] {argument});
        } catch (ClassCastException e) {
            throw new IllegalArgumentException("Programming error " 
                + e.toString());
        } catch (InvocationTargetException e) {
//            throw new IllegalArgumentException(e.getTargetException().getMessage()); 
            if (e.getTargetException() instanceof ArgumentException) {
                throw (ArgumentException) e.getTargetException().fillInStackTrace();
            } else {
                throw new IllegalArgumentException(e.getTargetException().getMessage());
            }
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException("Programming error "
                + e.toString());
        } catch (IllegalAccessException e) {
            throw new IllegalArgumentException("Programming error "
                + e.toString());
        } catch (InstantiationException e) {
            throw new IllegalArgumentException("Programming error "
                + e.toString());
        }

    }


    /**
     * Create an {@link Argument} out of an {@link TextInput} object.
     *
     * @return  created argument
     * @throws  ParsingException    if creation failed
     */
    public final Argument readArgument()
            throws ParsingException {
        final StringBuffer buffer = new StringBuffer();
        String key = null;
        char c;
        int begin = input.getPosition();
        input.skipWhiteSpace();
        if (input.isEmpty()) {
            input.setPosition(begin);
            throw new ParsingException(
                ModuleConstants.CODE_PARSER_ARGUMENT_EXPECTED,
                ModuleConstants.ARGUMENT_EXPECTED,
                input.getRow(),
                input.getColumn(),
                input.getPosition(),
                input.getLine(), input.getLocalAddress());
        }

        c = input.getChar();
        begin = input.getPosition();
        if (c == ')') {
            throw new ParsingException(
                ModuleConstants.CODE_PARSER_ARGUMENT_EXPECTED,
                ModuleConstants.ARGUMENT_EXPECTED,
                input.getRow(),
                input.getColumn(),
                input.getPosition(),
                input.getLine(), input.getLocalAddress());
        } else if (Character.isUpperCase(c)) {
            key = input.readLetterDigitString();
            Class cl = (Class) string2class.get(key);
            if (cl != null) {
                try {
                    final Argument argument;
                    if (cl == Module.class) {
                        try {
                            argument = new Module(readArgumentList(),
                            new ModuleAddress(input.getAddress()));
                        } catch (IOException e) {
                            throw new ArgumentException(
                                ModuleConstants.CODE_METHOD_ARGUMENT_CHECK_FAILED,
                                ModuleConstants.MALFORMED_MODULE_ADDRESS
                                + input.getAddress());
                        }
                    } else {
                        argument = createArgument(cl, readArgumentList());
                    }
                    final int end = input.getPosition();
                    parsing.put(argument, new Position(begin, end));
                    return argument;
                } catch (ArgumentException e) {
                    final Argument argument = e.getArgument();
                    if (argument != null) {
                        try {
                            final Position pos = parsing.get(argument);
                            if (e.isBegin()) {
                                input.setPosition(pos.getBegin());
                            } else {
                                input.setPosition(pos.getEnd());
                            }
                            throw new ParsingException(e.getCode(),
                                e.getDescription(),
                                input.getRow(),
                                input.getColumn(),
                                input.getPosition(),
                                input.getLine(), input.getLocalAddress());
                        } catch (IllegalArgumentException ex) {
                            // ignore and continue
                        }
                    }
                    input.setPosition(begin);
                    throw new ParsingException(e.getCode(),
                        e.getDescription(),
                        input.getRow(),
                        input.getColumn(),
                        input.getPosition(),
                        input.getLine(), input.getLocalAddress());
                }
            } else {
                input.setPosition(begin);
                throw new ParsingException(
                    ModuleConstants.CODE_PARSER_UNKNOWN_OPERATOR,
                    ModuleConstants.UNKNOWN_OPERATOR + Utility.quote(key),
                    input.getRow(),
                    input.getColumn(),
                    input.getPosition(),
                    input.getLine(), input.getLocalAddress());
            }
        } else if (Character.isDigit(c)) {
            try {
                final Counter counter = new Counter(input.readCounter());
                final int end = input.getPosition();
                parsing.put(counter, new Position(begin, end));
                return counter;
            } catch (IllegalArgumentException e) {
                throw new ParsingException(ModuleConstants
                    .CODE_PARSER_WRONG_ATOM_FORMAT,
                    e.getMessage(),
                    input.getRow(),
                    input.getColumn(),
                    input.getPosition(),
                    input.getLine(), input.getLocalAddress());
            } catch (ArgumentException e) {
                throw new ParsingException(e.getCode(),
                    e.getDescription(),
                    input.getRow(),
                    input.getColumn(),
                    input.getPosition(),
                    input.getLine(), input.getLocalAddress());
            }
        } else if (c == '\"') {
            try {
                final Text text = new Text(input.readQuoted());
                final int end = input.getPosition();
                parsing.put(text, new Position(begin, end));
                return text;
            } catch (IllegalArgumentException e) {
                throw new ParsingException(ModuleConstants
                    .CODE_PARSER_WRONG_ATOM_FORMAT,
                    e.getMessage(),
                    input.getRow(),
                    input.getColumn(),
                    input.getPosition(),
                    input.getLine(), input.getLocalAddress());
            }
        } else if (c == '(') {
                throw new ParsingException(ModuleConstants
                    .CODE_PARSER_ARGUMENT_EXPECTED,
                    ModuleConstants.ARGUMENT_EXPECTED,
                    input.getRow(),
                    input.getColumn(),
                    input.getPosition(),
                    input.getLine(), input.getLocalAddress());
        } else if (c == '.') {
                throw new ParsingException(ModuleConstants
                    .CODE_PARSER_ARGUMENT_EXPECTED,
                    ModuleConstants.ARGUMENT_EXPECTED,
                    input.getRow(),
                    input.getColumn(),
                    input.getPosition(),
                    input.getLine(), input.getLocalAddress());
        } else {
            throw new ParsingException(ModuleConstants
                .CODE_PARSER_UNKNOWN_ARGUMENT_CHAR,
                ModuleConstants.UNKNOWN_ARGUMENT_CHAR + "\"" + c + "\"",
                input.getRow(),
                input.getColumn(),
                input.getPosition(),
                input.getLine(), input.getLocalAddress());
        }
    }


    /**
     * Create an argument array out of an {@link TextInput} object.
     *
     * @return  created arguments
     * @throws  ParsingException    if creation failed
     */
    private final Argument[] readArgumentList()
            throws ParsingException {

        final int begin = input.getPosition();
        input.skipWhiteSpace();
        if (input.isEmpty() || input.getChar() != '(') {
            if (input.isEmpty()) {
                input.setPosition(begin);   // for showing correct position
            }
            throw new ParsingException(ModuleConstants
                .CODE_PARSER_NO_BEGIN_ARGUMENT_LIST,
                ModuleConstants.NO_BEGIN_ARGUMENT_LIST,
                input.getRow(),
                input.getColumn(),
                input.getPosition(),
                input.getLine(), input.getLocalAddress());

        }
        input.readChar();       // read '('

        final List argumentList = new ArrayList();
        int end = input.getPosition();
        input.skipWhiteSpace();
        if (!input.isEmpty() && ')' != input.getChar()) {
            do {
                argumentList.add(readArgument());
                end = input.getPosition();
                input.skipWhiteSpace();
                if (input.isEmpty() || ')' == input.getChar()) {
                    break;
                }
                if (input.getChar() == ',') {
                    input.readChar();
                } else {
                    throw new ParsingException(ModuleConstants
                        .CODE_PARSER_NO_ARGUMENT_SEPARATOR,
                        ModuleConstants.NO_ARGUMENT_SEPARATOR,
                        input.getRow(),
                        input.getColumn(),
                        input.getPosition(),
                        input.getLine(), input.getLocalAddress());
                }
            } while (true);
        }

        input.skipWhiteSpace();

        if (input.isEmpty() || ')' != input.getChar()) {
            if (input.isEmpty()) {
                input.setPosition(end);   // for showing correct position
            }
            throw new ParsingException(ModuleConstants
                .CODE_PARSER_NO_END_ARGUMENT_LIST,
                ModuleConstants.NO_END_ARGUMENT_LIST,
                input.getRow(),
                input.getColumn(),
                input.getPosition(),
                input.getLine(), input.getLocalAddress());
        }
        input.readChar();       // read ')'

        return (Argument[]) argumentList.toArray(
            new Argument[argumentList.size()]);
    }


    /**
     * Write an {@link Argument} to an {@link Output} object.
     *
     * @param   output   destination to write to
     * @param   argument this object will be written
     * @throws  IllegalArgumentException    if writing failed
     */
    public final static void writeArgument(final Output output,
            final Argument argument)
            throws IllegalArgumentException {
        final String function
            = (String) class2string.get(argument.getClass());
        if (function == null) {
            if (argument.getClass() == Counter.class) {
                output.write(((Counter) argument).getNumber());
                return;
            } else if (argument.getClass() == Text.class) {
                output.write(Utility.quote(((Text) argument).getText()));
                return;
            } else {
                throw new IllegalArgumentException(
                    ModuleConstants.UNKNOWN_ARGUMENT_CLASS
                    + argument.getClass());
            }
        }
        output.write(function);
        final int size = argument.getArgumentSize();
        output.write("(");
        for (int i = 0; i < size; i++) {
            if (i > 0) {
                output.write(", ");
            }
            writeArgument(output, argument.getArgument(i));
        }
        output.write(")");

    }


    /**
     * Write an {@link Argument} to an <code>String</code>.
     *
     * @param   argument this object will be written
     * @return  created string
     * @throws  IllegalArgumentException    if writing failed
     */
    public final static String writeArgument(
            final Argument argument)
            throws IllegalArgumentException {
        final StringBuffer buffer = new StringBuffer();
        writeArgument(new Output(buffer), argument);
        return buffer.toString();
    }


}
